/*
 * Machine-dependent part of uOS for: Atmel AVR, GCC.
 */
#include <stdlib.h>
#include <assert.h>
#include "sys/uos.h"
#include "sys/internal.h"
#include "sys/machdep.h"

/*
 * Clear bit in far register, using r30.
 */
#define clearb_r30(bit, port) do {			\
		asm volatile (				\
		"in r30, %0" "\n"			\
		"andi r30, %1" "\n"			\
		"out %0, r30"				\
		: /* no outputs */			\
		: "I" ((char)(port)),			\
		  "M" (0xff & ~(1 << (bit))));		\
	} while (0)

#define SAVE_REGS() asm volatile (\
	"push	r29\n"\
	"push	r28\n"\
	"push	r27\n"\
	"push	r26\n"\
	"push	r25\n"\
	"push	r24\n"\
	"push	r23\n"\
	"push	r22\n"\
	"push	r21\n"\
	"push	r20\n"\
	"push	r19\n"\
	"push	r18\n"\
	"push	r17\n"\
	"push	r16\n"\
	"push	r15\n"\
	"push	r14\n"\
	"push	r13\n"\
	"push	r12\n"\
	"push	r11\n"\
	"push	r10\n"\
	"push	r9\n"\
	"push	r8\n"\
	"push	r7\n"\
	"push	r6\n"\
	"push	r5\n"\
	"push	r4\n"\
	"push	r3\n"\
	"push	r2\n"\
	"push	r1\n"\
	"push	r0\n"\
	"clr	__zero_reg__\n")

/*
 * Perform the task switch.
 * The call is performed via the assembler label,
 * to skip the function prologue, generated by the compiler.
 */
void
_avr_switch (void)
{
	/* Skip function prologue. */
	asm volatile (
"avr_task_switch: .globl avr_task_switch\n"
	"push	r31\n"
	"in	r31, __SREG__\n"
	"cli\n"
	"push	r31\n"
	"push	r30\n");
	SAVE_REGS();

	/* Perform the context switch to the most priority task. */
	asm volatile (
"switch_task:");
	/* Save current task stack. */
	task_current->stack_pointer = (void*) inw (SP);

	/* Compute new active task. */
	task_policy ();

	/* Switch to the new task. */
	outw ((short) task_current->stack_pointer, SP);

	/* Restore registers. */
	asm volatile (
	"pop	r0\n"
	"pop	r1\n"
	"pop	r2\n"
	"pop	r3\n"
	"pop	r4\n"
	"pop	r5\n"
	"pop	r6\n"
	"pop	r7\n"
	"pop	r8\n"
	"pop	r9\n"
	"pop	r10\n"
	"pop	r11\n"
	"pop	r12\n"
	"pop	r13\n"
	"pop	r14\n"
	"pop	r15\n"
	"pop	r16\n"
	"pop	r17\n"
	"pop	r18\n"
	"pop	r19\n"
	"pop	r20\n"
	"pop	r21\n"
	"pop	r22\n"
	"pop	r23\n"
	"pop	r24\n"
	"pop	r25\n"
	"pop	r26\n"
	"pop	r27\n"
	"pop	r28\n"
	"pop	r29\n"
	"pop	r30\n"
	"pop	r31\n"
	"sbrc	r31, 7\n"		/* test I flag */
	"rjmp	enable_interrupts\n"
	"out	__SREG__, r31\n"
	"pop	r31\n"
	"ret\n"				/* exit with interrupts disabled */
"enable_interrupts:\n"
	"andi	r31, 0x7f\n"		/* clear I flag */
	"out	__SREG__, r31\n"
	"pop	r31\n"
	"reti\n");			/* exit with interrupts enabled */
}

#if FLASHEND > 0x2000
#define GOTO "jmp "
#else
#define GOTO "rjmp "
#endif

/*
 * The common part of the interrupt handler,
 * to minimize the code size.
 */
void
_avr_intr (void)
{
	register lock_t **lock asm ("r30") = uos_handler;

	asm volatile (
"handle_interrupt:");
	SAVE_REGS();

	/* Signal the interrupt handler, if any. */
	lock_activate (*lock, 0);

	asm volatile (GOTO "switch_task");
}

/*
 * The interrupt handler pattern.
 */
#define HANDLE(n,name,mask)\
void \
_intr##n (void) \
{\
	asm volatile (#name": .globl "#name"\n"\
	"push	r31\n"\
	"in	r31, __SREG__\n"\
	"ori	r31, 0x80\n"		/* set I flag on exit */\
	"push	r31\n"\
	"push	r30\n");\
	mask;				/* disable the irq, avoiding loops */\
	asm volatile (GOTO "handle_interrupt" : : "z" (uos_handler + n));\
}

#ifdef INT0_vect
	HANDLE(0, _interrupt0_, clearb_r30 (0, EIMSK))
#endif
#ifdef INT1_vect
	HANDLE(1, _interrupt1_, clearb_r30 (1, EIMSK))
#endif
#ifdef INT2_vect
	HANDLE(2, _interrupt2_, clearb_r30 (2, EIMSK))
#endif
#ifdef INT3_vect
	HANDLE(3, _interrupt3_, clearb_r30 (3, EIMSK))
#endif
#ifdef INT4_vect
	HANDLE(4, _interrupt4_, clearb_r30 (4, EIMSK))
#endif
#ifdef INT5_vect
	HANDLE(5, _interrupt5_, clearb_r30 (5, EIMSK))
#endif
#ifdef INT6_vect
	HANDLE(6, _interrupt6_, clearb_r30 (6, EIMSK))
#endif
#ifdef INT7_vect
	HANDLE(7, _interrupt7_, clearb_r30 (7, EIMSK))
#endif
#ifdef TIMER2_COMP_vect
	HANDLE(8, _output_compare2_, clearb_r30 (7, TIMSK))
#endif
#ifdef TIMER2_OVF_vect
	HANDLE(9, _overflow2_, clearb_r30 (6, TIMSK))
#endif
#ifdef TIMER1_CAPT_vect
	HANDLE(10, _input_capture1_, clearb_r30 (5, TIMSK))
#endif
#ifdef TIMER1_COMPA_vect
	HANDLE(11, _output_compare1a_, clearb_r30 (4, TIMSK))
#endif
#ifdef TIMER1_COMPB_vect
	HANDLE(12, _output_compare1b_, clearb_r30 (3, TIMSK))
#endif
#ifdef TIMER1_OVF_vect
	HANDLE(13, _overflow1_, clearb_r30 (2, TIMSK))
#endif
#ifdef TIMER0_COMP_vect
	HANDLE(14, _output_compare0_, clearb_r30 (1, TIMSK))
#endif
#ifdef TIMER0_OVF_vect
	HANDLE(15, _overflow0_, clearb_r30 (0, TIMSK))
#endif
#ifdef SPI_STC_vect
	HANDLE(16, _spi_, clearb_const (SPIE, SPCR))
#endif
#ifdef UART_RX_vect
	HANDLE(17, _uart_recv_, clearb_const (RXCIE, UCR))
#endif
#ifdef UART_UDRE_vect
	HANDLE(18, _uart_data_, clearb_const (UDRIE, UCR))
#endif
#ifdef UART_TX_vect
	HANDLE(19, _uart_trans_, clearb_const (TXCIE, UCR))
#endif
#ifdef ADC_vect
	HANDLE(20, _adc_, clearb_const (ADIE, ADCSR))
#endif
#ifdef EE_RDY_vect
	HANDLE(21, _eeprom_ready_, clearb_const (EERIE, EECR))
#endif
#ifdef ANA_COMP_vect
	/* Allow gdb-stub to grab the analog comparator interrupt. */
	asm (".weak _comparator_");
	HANDLE(22, _comparator_, clearb_const (ACIE, ACSR))
#endif
#ifdef UART1_RX_vect
	HANDLE(23, _uart1_recv_, clearb_const (RXCIE1, UCSR1B))
#endif
#ifdef UART1_UDRE_vect
	HANDLE(24, _uart1_data_, clearb_const (UDRIE1, UCSR1B))
#endif
#ifdef UART1_TX_vect
	HANDLE(25, _uart1_trans_, clearb_const (TXCIE1, UCSR1B))
#endif

/*
 * Allow the given hardware interrupt,
 * unmasking it in the interrupt controller.
 */
void
avr_intr_allow (unsigned char irq)
{
	switch (irq) {
#ifdef INT0_vect			/* External interrupts. */
	case 0: setb (0, EIMSK); break;
#endif
#ifdef INT1_vect
	case 1: setb (1, EIMSK); break;
#endif
#ifdef INT2_vect
	case 2: setb (2, EIMSK); break;
#endif
#ifdef INT3_vect
	case 3: setb (3, EIMSK); break;
#endif
#ifdef INT4_vect
	case 4: setb (4, EIMSK); break;
#endif
#ifdef INT5_vect
	case 5: setb (5, EIMSK); break;
#endif
#ifdef INT6_vect
	case 6: setb (6, EIMSK); break;
#endif
#ifdef INT7_vect
	case 7: setb (7, EIMSK); break;
#endif
#ifdef TIMER2_COMP_vect			/* Timer/counter interrupts. */
	case 8: setb (7, TIMSK); break;
#endif
#ifdef TIMER2_OVF_vect
	case 9: setb (6, TIMSK); break;
#endif
#ifdef TIMER1_CAPT_vect
	case 10: setb (5, TIMSK); break;
#endif
#ifdef TIMER1_COMPA_vect
	case 11: setb (4, TIMSK); break;
#endif
#ifdef TIMER1_COMPB_vect
	case 12: setb (3, TIMSK); break;
#endif
#ifdef TIMER1_OVF_vect
	case 13: setb (2, TIMSK); break;
#endif
#ifdef TIMER0_COMP_vect
	case 14: setb (1, TIMSK); break;
#endif
#ifdef TIMER0_OVF_vect
	case 15: setb (0, TIMSK); break;
#endif
#ifdef SPI_STC_vect			/* SPI event */
	case 16: setb (SPIE, SPCR); break;
#endif
#ifdef UART_RX_vect			/* UART receive complete */
	case 17: setb (RXCIE, UCR); break;
#endif
#ifdef UART_UDRE_vect			/* UART transmitter empty */
	case 18: setb (UDRIE, UCR); break;
#endif
#ifdef UART_TX_vect			/* UART transmit complete */
	case 19: setb (TXCIE, UCR); break;
#endif
#ifdef ADC_vect				/* Analog-digital conversion complete */
	case 20: setb (ADIE, ADCSR); break;
#endif
#ifdef EE_RDY_vect			/* EEPROM ready */
	case 21: setb (EERIE, EECR); break;
#endif
#ifdef ANA_COMP_vect			/* Analog comparator event */
	case 22: setb (ACIE, ACSR); break;
#endif
#ifdef UART1_RX_vect			/* UART 1 receive complete */
	case 23: setb (RXCIE1, UCSR1B); break;
#endif
#ifdef UART1_UDRE_vect			/* UART 1 transmitter empty */
	case 24: setb (UDRIE1, UCSR1B); break;
#endif
#ifdef UART1_TX_vect			/* UART 1 transmit complete */
	case 25: setb (TXCIE1, UCSR1B); break;
#endif
	}
}

/*
 * Build the initial task's stack frame.
 * Arguments:
 * t	- the task object
 * func	- the task function to call
 * arg	- the function argument
 * sp	- the pointer to (end of) stack space
 */
void
avr_build_stack_frame (task_t *t, unsigned short func, unsigned short arg,
	char *sp)
{
	*sp-- = 0;			/* caller address low... */
	*sp-- = 0;			/* ...high */
	*sp-- = func;			/* callee address low... */
	*sp-- = func >> 8;		/* ...high */
	*sp-- = 0;			/* r31 */
	*sp-- = 0x80;			/* SREG - enable interrupts */
	*sp-- = 0;			/* r30 */
	*sp-- = 0;			/* r29 */
	*sp-- = 0;			/* r28 */
	*sp-- = 0;			/* r27 */
	*sp-- = 0;			/* r26 */
	*sp-- = arg >> 8;		/* r25 - task argument high... */
	*sp-- = arg;			/* r24 ...low */
	*sp-- = 0;			/* r23 */
	*sp-- = 0;			/* r22 */
	*sp-- = 0;			/* r21 */
	*sp-- = 0;			/* r20 */
	*sp-- = 0;			/* r19 */
	*sp-- = 0;			/* r18 */
	*sp-- = 0;			/* r17 */
	*sp-- = 0;			/* r16 */
	*sp-- = 0;			/* r15 */
	*sp-- = 0;			/* r14 */
	*sp-- = 0;			/* r13 */
	*sp-- = 0;			/* r12 */
	*sp-- = 0;			/* r11 */
	*sp-- = 0;			/* r10 */
	*sp-- = 0;			/* r9 */
	*sp-- = 0;			/* r8 */
	*sp-- = 0;			/* r7 */
	*sp-- = 0;			/* r6 */
	*sp-- = 0;			/* r5 */
	*sp-- = 0;			/* r4 */
	*sp-- = 0;			/* r3 */
	*sp-- = 0;			/* r2 */
	*sp-- = 0;			/* r1 */
	*sp-- = 0;			/* r0 */
	t->stack_pointer = (void*) sp;
}
