#include "sys/uos.h"
#include "sys/lib.h"
#include "sys/uart.h"

#if __AVR__
#   include <machine/io.h>
#   define RECEIVE_IRQ		17	/* UART receive complete */
#   define TRANSMIT_IRQ		18	/* UART transmit complete */
#endif

/*
 * Wait for transmitter to finish sending the current byte.
 */
static void
uart_wait_for_transmit (uart_t *u)
{
	for (;;) {
#ifndef NDEBUG
		if (testb (RXC, USR)) {
			u->buffered_char = inb (UDR);
			if (u->buffered_char == 3) {
				u->buffered_char = -1;
				breakpoint ();
			}
		}
#endif
		/* Wait until transmit data empty. */
		if (testb (UDRE, USR))
			return;
		lock_wait (&u->transmitter);
	}
}

/*
 * Send a byte to the UART transmitter.
 */
void
uart_putchar (uart_t *u, char c)
{
	lock_acquire (&u->transmitter);

	/* Wait for transmitter idle. */
	uart_wait_for_transmit (u);
again:
	/* Send byte, enable interrupts. */
	outb (c, UDR);
	setb (UDRIE, UCR);

	/* Wait for transmitter idle. */
	uart_wait_for_transmit (u);

	if (c == '\n') {
		c = '\r';
		goto again;
	}

	/* Disable transmit interrupt. */
	clearb (UDRIE, UCR);
	lock_release (&u->transmitter);
}

/*
 * Wait for the byte to be received and return it.
 */
unsigned char
uart_getchar (uart_t *u)
{
	unsigned char c;

	lock_acquire (&u->receiver);

	/* Wait until receive data available. */
	while (u->buffered_char < 0)
		lock_wait (&u->receiver);

	c = u->buffered_char;
	u->buffered_char = -1;
	lock_release (&u->receiver);
	return c;
}

/*
 * System uart task.
 */
static void
main_uart (void *arg)
{
	uart_t *u = arg;

	/* Enable receiver. */
	lock_acquire_irq (&u->receiver, RECEIVE_IRQ);
	setb (RXEN, UCR);
	setb (RXCIE, UCR);
	u->buffered_char = -1;
	for (;;) {
		lock_wait (&u->receiver);
		if (testb (RXC, USR)) {
			u->buffered_char = inb (UDR);
#ifndef NDEBUG
			if (u->buffered_char == 3) {
				breakpoint ();
				u->buffered_char = -1;
			}
#endif
		}
	}
}

void
uart_init (uart_t *u, int prio, unsigned long baud)
{
	/* Enable transmitter. */
	lock_acquire_irq (&u->transmitter, TRANSMIT_IRQ);
	outb (((int) (KHZ * 1000L / baud) + 8) / 16 - 1, UBRR);
	setb (TXEN, UCR);
	lock_release (&u->transmitter);

	/* Create uart receive task. */
	task_create (main_uart, u, prio, u->stack, sizeof (u->stack), 0);
}
