#include <stdlib.h>
#include <assert.h>
#include "sys/uos.h"
#include "sys/machdep.h"
#include "sys/internal.h"

/*
 * Utility functions.
 * Must be called with interrupts disabled.
 */
void
lock_enqueue (lock_list_t *list, lock_t *elem)
{
	if (! list->head) {
		assert (list->tail == 0);
		elem->next = 0;
		elem->prev = 0;
		list->head = elem;
		list->tail = elem;
	} else {
		assert (list->tail != 0);
		elem->next = list->head;
		elem->prev = 0;
		list->head->prev = elem;
		list->head = elem;
	}
}

void
lock_dequeue (lock_list_t *list, lock_t *elem)
{
	assert (list->head != 0);
	assert (list->tail != 0);
	if (list->head == elem) {
		list->head = elem->next;
		assert (elem->prev == 0);
	} else {
		assert (elem->prev != 0);
		elem->prev->next = elem->next;
	}
	if (list->tail == elem) {
		list->tail = elem->prev;
		assert (elem->next == 0);
	} else {
		assert (elem->next != 0);
		elem->next->prev = elem->prev;
	}
}

/*
 * Get the lock. The calling task would block
 * waiting for the lock to be released by somebody.
 * In the case the lock has associated IRQ number,
 * after acquiring the lock the IRQ will be disabled.
 */
void
lock_acquire (lock_t *m)
{
	int x;

	MACHDEP_INTR_DISABLE (&x);
	assert (STACK_GUARD (task_current));
	while (m->master && m->master != task_current) {
		/* Monitor is locked, block the task. */
		assert (task_current->lock == 0);
		task_current->lock = m;
		task_dequeue (&task_active, task_current);
		task_enqueue (&m->slaves, task_current);
		MACHDEP_TASK_SWITCH ();
	}
	if (! m->master) {
		assert (m->slaves.head == 0);
		m->master = task_current;
		lock_enqueue (&task_current->slaves, m);
	}
	if (m->irq)
		m->context = x;
	else
		MACHDEP_INTR_RESTORE (x);
}

/*
 * Release the lock. All tasks waiting for the lock
 * are unblocked, possibly causing task switch.
 * In the case the lock has associated IRQ number,
 * the IRQ will be enabled.
 */
void
lock_release (lock_t *m)
{
	int x;

	assert (STACK_GUARD (task_current));
	if (m->irq)
		x = m->context;
	else
		MACHDEP_INTR_DISABLE (&x);
	assert (m->master != 0);
	lock_dequeue (&m->master->slaves, m);
	m->master = 0;
	if (! m->slaves.head) {
		MACHDEP_INTR_RESTORE (x);
		return;
	}
	do {
		task_t *t = m->slaves.head;
		assert (t->lock == m);
		t->lock = 0;
		task_dequeue (&m->slaves, t);
		task_enqueue (&task_active, t);
	} while (m->slaves.head);
	MACHDEP_INTR_RESTORE (x);
	MACHDEP_TASK_SWITCH ();
}
