#include <stdlib.h>
#include <assert.h>
#include "sys/uos.h"
#include "sys/machdep.h"
#include "sys/internal.h"

task_list_t task_active;			/* list of tasks ready to run */
task_t *task_current;				/* current running task */
static task_t task_idle;			/* background system task */
lock_t *uos_handler [MACHDEP_INTERRUPTS];	/* interrupt handlers */

/*
 * Utility functions.
 * Must be called with interrupts disabled.
 */
void
task_enqueue (task_list_t *list, task_t *elem)
{
	if (! list->head) {
		assert (list->tail == 0);
		elem->next = 0;
		elem->prev = 0;
		list->head = elem;
		list->tail = elem;
	} else {
		assert (list->tail != 0);
		elem->next = list->head;
		elem->prev = 0;
		list->head->prev = elem;
		list->head = elem;
	}
}

void
task_dequeue (task_list_t *list, task_t *elem)
{
	assert (list->head != 0);
	assert (list->tail != 0);
	if (list->head == elem) {
		list->head = elem->next;
		assert (elem->prev == 0);
	} else {
		assert (elem->prev != 0);
		elem->prev->next = elem->next;
	}
	if (list->tail == elem) {
		list->tail = elem->prev;
		assert (elem->next == 0);
	} else {
		assert (elem->next != 0);
		elem->next->prev = elem->prev;
	}
}

/*
 * Activate all waiters of the lock.
 */
void
lock_activate (lock_t *m, void *message)
{
	task_t *t;

	if (! m)
		return;
	for (;;) {
		t = m->waiters.head;
		if (! t)
			return;
		assert (t->wait == m);
		t->wait = 0;
		t->message = message;
		task_dequeue (&m->waiters, t);
		task_enqueue (&task_active, t);
	}
}

/*
 * Find the task with maximum priority,
 * and put it to the head of queue.
 * The core of the scheduler policy.
 */
task_t *
task_policy (void)
{
	int maxprio = 0;
	task_t *t, *r;

	assert (STACK_GUARD (task_current));
	r = &task_idle;
	for (t=task_active.head; t; t=t->next) {
		if (t->prio >= maxprio) {
			r = t;
			maxprio = t->prio;
		}
	}
	task_current = r;
	return r;
}

/*
 * Call user initialization routine uos_init(),
 * then create the idle task, and run the OS.
 * The idle task uses the default system stack.
 */
int
main (void)
{
	/* Create the idle task. */
	task_idle.stack[0] = STACK_MAGIC;
	task_current = &task_idle;
	task_enqueue (&task_active, &task_idle);

	/* Create user tasks. */
	uos_init ();

	/* Switch to the most priority task. */
	MACHDEP_TASK_SWITCH ();

	/* Idle task activity. */
	for (;;) {
#ifdef MACHDEP_IDLE
		MACHDEP_IDLE ();
#endif
	}
}
