/*
 * Scaled down version of printf(3).
 *
 * Two additional formats:
 *
 * The format %b is supported to decode error registers.
 * Its usage is:
 *
 *	printf("reg=%b\n", regval, "<base><arg>*");
 *
 * where <base> is the output base expressed as a control character, e.g.
 * \10 gives octal; \20 gives hex.  Each arg is a sequence of characters,
 * the first of which gives the bit number to be inspected (origin 1), and
 * the next characters (up to a control character, i.e. a character <= 32),
 * give the name of the register.  Thus:
 *
 *	kvprintf("reg=%b\n", 3, "\10\2BITTWO\1BITONE\n");
 *
 * would produce output:
 *
 *	reg=3<BITTWO,BITONE>
 *
 * The format %D -- Hexdump, takes a pointer. Sharp flag - use `:' as
 * a separator, instead of a space. For example:
 *
 *	("%6D", ptr)       -> XX XX XX XX XX XX
 *	("%#*D", len, ptr) -> XX:XX:XX:XX ...
 */
#include <sys/lib.h>

/* Max number conversion buffer length: a long in base 2, plus NUL byte. */
#define MAXNBUF	(sizeof(long) * 8 + 1)

static char *ksprintn (char *buf, unsigned long v, unsigned char base,
	unsigned char *lp);
static char mkhex (char ch);

int
doprnt (void (*func)(void*, char), void *arg, char const *fmt, va_list ap)
{
#define PUTC(c) { (*func) (arg, (c)); ++retval; }
	char nbuf [MAXNBUF], padding, *p, *q;
	unsigned char ch, base, lflag, ladjust, sharpflag, neg, sign, dot, tmp;
	unsigned char *up;
	char *pp;
	int n, width, dwidth, retval;
	unsigned long ul;

	if (! func)
		return 0;
	if (! fmt)
		fmt = CONST("(null)\n");

	retval = 0;
	for (;;) {
		padding = ' ';
		width = 0;
		while ((ch = FETCH (fmt++)) != '%') {
			if (! ch)
				return retval;
			PUTC (ch);
		}
		lflag = 0; ladjust = 0; sharpflag = 0; neg = 0;
		sign = 0; dot = 0; dwidth = 0;
reswitch:	switch (ch = FETCH (fmt++)) {
		case '.':
			dot = 1;
			goto reswitch;

		case '#':
			sharpflag = 1;
			goto reswitch;

		case '+':
			sign = 1;
			goto reswitch;

		case '-':
			ladjust = 1;
			goto reswitch;

		case '%':
			PUTC (ch);
			break;

		case '*':
			if (! dot) {
				width = va_arg (ap, int);
				if (width < 0) {
					ladjust = !ladjust;
					width = -width;
				}
			} else {
				dwidth = va_arg (ap, int);
			}
			goto reswitch;

		case '0':
			if (! dot) {
				padding = '0';
				goto reswitch;
			}
		case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			for (n=0; ; ++fmt) {
				n = n * 10 + ch - '0';
				ch = FETCH (fmt);
				if (ch < '0' || ch > '9')
					break;
			}
			if (dot)
				dwidth = n;
			else
				width = n;
			goto reswitch;

		case 'b':
			ul = va_arg (ap, int);
			p = va_arg (ap, char*);
			q = ksprintn (nbuf, ul, *p++, 0);
			while (*q)
				PUTC (*q--);

			if (! ul)
				break;
			tmp = 0;
			while (*p) {
				n = *p++;
				if ((char) (ul >> (n-1)) & 1) {
					PUTC (tmp ? ',' : '<');
					for (; (n = *p) > ' '; ++p)
						PUTC (n);
					tmp = 1;
				} else
					while (*p > ' ')
						++p;
			}
			if (tmp)
				PUTC ('>');
			break;

		case 'c':
			PUTC (va_arg (ap, int));
			break;

		case 'D':
			up = va_arg (ap, unsigned char*);
			if (! width)
				width = 16;
			if (sharpflag)
				padding = ':';
			while (width--) {
				ch = *up++;
				PUTC (mkhex (ch >> 4));
				PUTC (mkhex (ch));
				if (width)
					PUTC (padding);
			}
			break;

		case 'd':
			ul = lflag ? va_arg (ap, long) : va_arg (ap, int);
			sign = 1;
			base = 10;
			goto number;

		case 'l':
			lflag = 1;
			goto reswitch;

		case 'o':
			ul = lflag ? va_arg (ap, unsigned long) :
				va_arg (ap, unsigned int);
			base = 8;
			goto nosign;

		case 'p':
			ul = (unsigned short) va_arg (ap, void*);
			base = 16;
			sharpflag = (width == 0);
			goto nosign;

		case 'n':
		case 'r':
			ul = lflag ? va_arg (ap, unsigned long) :
				sign ? (unsigned long) va_arg (ap, int) :
				va_arg (ap, unsigned int);
			base = 10;
			goto number;

		case 's':
			p = va_arg (ap, char*);
			if (! p)
				p = "(null)";
			if (! dot)
				n = strlen (p);
			else
				for (n=0; n<dwidth && p[n]; n++)
					continue;

			width -= n;

			if (! ladjust && width > 0)
				while (width--)
					PUTC (padding);
			while (n--)
				PUTC (*p++);
			if (ladjust && width > 0)
				while (width--)
					PUTC (padding);
			break;

		case 'S':
			pp = va_arg (ap, char*);
			if (! pp)
				pp = CONST("(null)");
			n = 0;
			if (! dot)
				while (FETCH (pp+n))
					++n;
			else
				while (n<dwidth && FETCH (pp+n))
					++n;

			width -= n;

			if (! ladjust && width > 0)
				while (width--)
					PUTC (padding);
			while (n--)
				PUTC (FETCH (pp++));
			if (ladjust && width > 0)
				while (width--)
					PUTC (padding);
			break;

		case 'u':
			ul = lflag ? va_arg (ap, unsigned long) :
				va_arg (ap, unsigned int);
			base = 10;
			goto nosign;

		case 'x':
			ul = lflag ? va_arg (ap, unsigned long) :
				va_arg (ap, unsigned int);
			base = 16;
			goto nosign;

		case 'z':
			ul = lflag ? va_arg (ap, unsigned long) :
				sign ? (unsigned long) va_arg (ap, int) :
				va_arg (ap, unsigned int);
			base = 16;
			goto number;
nosign:			sign = 0;
number:			if (sign && (long) ul < 0L) {
				neg = 1;
				ul = -(long) ul;
			}
			p = ksprintn (nbuf, ul, base, &tmp);
			if (sharpflag && ul != 0) {
				if (base == 8)
					tmp++;
				else if (base == 16)
					tmp += 2;
			}
			if (neg)
				tmp++;

			if (! ladjust && width && (width -= tmp) > 0)
				while (width--)
					PUTC (padding);
			if (neg)
				PUTC ('-');
			if (sharpflag && ul != 0) {
				if (base == 8) {
					PUTC ('0');
				} else if (base == 16) {
					PUTC ('0');
					PUTC ('x');
				}
			}

			while (*p)
				PUTC (*p--);

			if (ladjust && width && (width -= tmp) > 0)
				while (width--)
					PUTC (padding);
			break;

		default:
			PUTC ('%');
			if (lflag)
				PUTC ('l');
			PUTC (ch);
			break;
		}
	}
}

/*
 * Put a NUL-terminated ASCII number (base <= 16) in a buffer in reverse
 * order; return an optional length and a pointer to the last character
 * written in the buffer (i.e., the first character of the string).
 * The buffer pointed to by `nbuf' must have length >= MAXNBUF.
 */
static char *
ksprintn (char *nbuf, unsigned long ul, unsigned char base, unsigned char *lenp)
{
	char *p;

	p = nbuf;
	*p = 0;
	do {
		*++p = mkhex (ul % base);
	} while (ul /= base);
	if (lenp)
		*lenp = p - nbuf;
	return (p);
}

static char
mkhex (char ch)
{
	ch &= 15;
	if (ch > 9)
		return ch + 'a' - 10;
	return ch + '0';
}

#if 0
void
doprnt_test (void)
{
	printf (CONST("Test b: %b\n"), 55, "\20\7b7\6b6\5b5\4b4\3b3\2b2\1b1");
	printf (CONST("Test c: %c (Z)\n"), 'Z');
	printf (CONST("Test D: %D\n"), (char*)0, ":");
	printf (CONST("Test p: %p\n"), &doprnt);
	printf (CONST("Test d: %d (1234)\n"), 1234);
	printf (CONST("Test l: %l (12345678)\n"), 12345678L);
	printf (CONST("Test o: %o (01234)\n"), 01234);
	printf (CONST("Test n: %n (1234)\n"), 1234);
	printf (CONST("Test s: %s (abra)\n"), "abra");
	printf (CONST("Test S: %S (cadabra)\n"), CONST("cadabra"));
	printf (CONST("Test u: %u (-1)\n"), -1);
	printf (CONST("Test x: %x (0x1234)\n"), 0x1234);
	printf (CONST("Test z: %z (0x1234)\n"), 0x1234);
}
#endif
