/* System data structures. */
typedef struct _task_t task_t;
typedef struct _lock_t lock_t;
typedef struct _task_list_t task_list_t;
typedef struct _lock_list_t lock_list_t;

/* Task management. */
task_t *task_create (void (*func)(void*), void *arg, int prio,
	char *stack, int stacksz, lock_t *finalizer);
void task_delete (task_t*, void *message);
void *task_wait (task_t*);
void task_exit (void *message);
int task_stack_avail (char *stack);

/* Lock management. */
void lock_acquire (lock_t*);
void lock_acquire_irq (lock_t*, int irq);
int lock_try (lock_t*);
void lock_release (lock_t*);
void lock_release_irq (lock_t*);
void lock_signal (lock_t*, void *message);
void *lock_wait (lock_t*);

/* Utility calls. */
void uos_halt (void);

/* User-supplied startup routine. */
extern void uos_init (void);

/*
 * Queue data structures.
 */
struct _task_list_t {
	task_t *	head;
	task_t *	tail;
};

struct _lock_list_t {
	lock_t *	head;
	lock_t *	tail;
};

/*
 * ----------
 * | Task   |
 * | ~~~~   |
 * | lock ----> M
 * |	    |
 * | wait ----> M
 * |	    |
 * | slaves --> M -> M -> M...
 * ---------- <-/----/----/
 */
struct _task_t {
	task_t *	next;
	task_t *	prev;
	lock_t *	lock;		/* lock, blocking the task */
	lock_t *	wait;		/* lock, the task is waiting for */
	lock_t *	finalizer;	/* lock to signal the task finish */
	lock_list_t	slaves;		/* locks, acquired by task */
	void *		message;	/* wait() return value */
	void *		stack_pointer;
	int		prio;		/* task priority */
	int		context;	/* saved when waiting for the lock */
	char 		stack [1];
};

/*
 * ----------
 * | Lock   |
 * | ~~~~   |
 * | master --> T
 * |	    |
 * | waiters--> T -> T...
 * |	    |
 * | slaves --> T -> T -> T...
 * ---------- <-/----/----/
 */
struct _lock_t {
	lock_t *	next;
	lock_t *	prev;
	task_t *	master;		/* task, acquired the lock */
	task_list_t	waiters;	/* tasks, stopped on `wait' */
	task_list_t	slaves;		/* tasks, waiting for lock */
	int		irq;		/* irq(+1) associated with the lock */
	int		context;	/* saved when locking the irq */
};
